
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "xstd/xstd.h"

#include "xstd/h/iostream.h"
#include "xstd/h/strstream.h"

#include "xstd/Assert.h"
#include "xstd/String.h"
#include "xstd/gadgets.h"


/* StrBuf */

void StrBuf::alloc(int aSize) {
	theBuf = new char[theSize = aSize];
}

void StrBuf::set(const char *aBuf, int aSize) {
	Assert(!theUseLvl);
	//cerr << here << "new string: " << aSize << ' ' << aBuf << endl;
	delete[] theBuf;
	alloc(aSize);
	memcpy(theBuf, aBuf, aSize);
}

StrBuf *StrBuf::clone() const {
	StrBuf *sb = new StrBuf;
	sb->set(theBuf, theSize);
	return sb;
}

void StrBuf::realloc(int newSize) {
	if (newSize > theSize) {
		char *buf = theBuf;
		const int sz = theSize;
		theBuf = 0;
		alloc(newSize);
		if (buf)
			memcpy(theBuf, buf, sz);
		delete[] buf;
	}
}


/* String */

String::String(const char *s, int len): theBuf(0) {
	Assert(s);
	Assert(len >= 0);

	theBuf = new StrBuf;
	theBuf->alloc(len+1);
	if (len > 0)
		memcpy(theBuf->theBuf, s, len);
	theBuf->theBuf[len] = '\0';
	theBuf->use();
}

void String::set(const char *s) {
	if (theBuf)
		theBuf = theBuf->abandon();

	if (s) {
		theBuf = new StrBuf;
		theBuf->set(s, strlen(s)+1);
		theBuf->use();
	}
}

void String::set(const String &s) {
	if (theBuf == s.theBuf)
		return; // either "copy to self" or "0 = 0"

	if (theBuf)
		theBuf->abandon();

	if ((theBuf = s.theBuf))
		theBuf->use();
}

void String::alloc(int size) {
	if (theBuf)
		theBuf->abandon();
	theBuf = new StrBuf;
	theBuf->alloc(size);
	theBuf->use();
}

void String::stretch(int newSize) {
	if (size() < newSize) {
		if (theBuf)
			theBuf->realloc(newSize);
		else
			alloc(newSize);
	}
}

// clone a private copy if we are using shared memory
// (copy-on-write or lazy copying)
void String::aboutToMod() {
	if (theBuf && theBuf->theUseLvl > 1) {
		StrBuf *sb = theBuf;
		theBuf = sb->clone();
		theBuf->use();
		sb->abandon();
	}
}

const char *String::chr(char c) const {
	return theBuf ? strchr(str(), c) : 0;
}

const char *String::rchr(char c) const {
	return theBuf ? strrchr(str(), c) : 0;
}

const char *String::str(const char *s) const {
	return theBuf ? strstr(str(), s) : 0;
}

String String::operator ()(int min, int maxp1) const {
	maxp1 = Min(maxp1, len());
	const int len = maxp1 - min;
	Assert(len >= 0);
	return String(str()+min, len);

#if 0 /* old working code */
	StrBuf *sb = new StrBuf;
	maxp1 = Min(maxp1, len());
	const int len = maxp1 - min;
	Assert(len >= 0);
	sb->alloc(len+1);
	if (len)
		memcpy(sb->theBuf, str()+min, len);
	sb->theBuf[len] = '\0';
	return *sb;
#endif
}

bool String::operator ==(const String &s) const {
	return 
		theBuf && s.theBuf &&
		theBuf->theBuf && s.theBuf->theBuf &&
		!strcmp(theBuf->theBuf, s.theBuf->theBuf);
}

// adds 0 after appending if not already there
void String::append(const char *buf, int size) {
	Assert(buf || !size);

	aboutToMod();

	const bool addTerm = !size || buf[size-1] != (char)0;
	const int l = len();

	stretch(l + size + (int)addTerm);
	Assert(theBuf);
	if (size)
		memcpy(theBuf->theBuf+l, buf, size);
	theBuf->theBuf[len()] = '\0';
}

String &String::operator +=(const String &s) {
	append(s.str(), s.len());
	return *this;
}

String &String::operator +=(char c) {
	append(&c, 1);
	return *this;
}


/* operators on strings */

ostream &operator <<(ostream &os, const String &s) {
	if (s) {
		if (s.len() > 0)
			os.write(s.str(), s.len());
	} else
		os << "<none>";
	return os;
}

String operator +(const String &s1, const String &s2) {
	String s;
	s += s1;
	s += s2;
	return s;
}

String Stream2String(ostrstream &os) {
	const String s = os.str();
	ostrstreamFreeze(os, false);
	return s;
}


/* StrIter */

StrIter::StrIter(const String &aStr, char aDel): 
	theStr(aStr), theTokBeg(0), theTokEnd(0), theDel(aDel) {
	if (theStr) {
		theTokBeg = theTokEnd = theStr.str();
		moveEnd();
	}
}

String StrIter::token() const {
	String s;
	s.append(tokBeg(), tokLen());
	return s;
}

void StrIter::next() {
	if (*theTokEnd) {
		theTokBeg = ++theTokEnd;
		moveEnd();
	} else {
		theTokBeg = theTokEnd;
	}
}

void StrIter::moveEnd() {
	while (*theTokEnd && *theTokEnd != theDel)
		++theTokEnd;
}
